"""
MobileNet v3, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/mobilenetv3.py

Paper "Searching for MobileNetV3", https://arxiv.org/abs/1905.02244

Changes from original:
* Using nn.BatchNorm2d with eps 1e-5 instead of 1e-3
"""

# Reference license: BSD 3-Clause

from collections import OrderedDict
from collections.abc import Callable
from typing import Any
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import SqueezeExcitation

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import make_divisible


class InvertedResidualConfig:
    """
    Stores information listed at Tables 1 and 2 of the MobileNet v3 paper
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        expanded_channels: int,
        kernel_size: tuple[int, int],
        stride: tuple[int, int],
        padding: tuple[int, int],
        width_multi: float,
        use_se: bool,
        activation: Callable[..., nn.Module] = nn.ReLU,
    ):
        self.in_channels = make_divisible(in_channels * width_multi, 8)
        self.out_channels = make_divisible(out_channels * width_multi, 8)
        self.expanded_channels = make_divisible(expanded_channels * width_multi, 8)
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.use_se = use_se
        self.activation = activation


class InvertedResidual(nn.Module):
    def __init__(self, cnf: InvertedResidualConfig) -> None:
        super().__init__()
        if cnf.stride[0] == 1 and cnf.stride[1] == 1 and cnf.in_channels == cnf.out_channels:
            self.shortcut = True
        else:
            self.shortcut = False

        layers = []
        # Expand
        if cnf.expanded_channels != cnf.in_channels:
            layers.append(
                Conv2dNormActivation(
                    cnf.in_channels,
                    cnf.expanded_channels,
                    kernel_size=(1, 1),
                    stride=(1, 1),
                    padding=(0, 0),
                    bias=False,
                    activation_layer=cnf.activation,
                )
            )

        # Depthwise
        layers.append(
            Conv2dNormActivation(
                cnf.expanded_channels,
                cnf.expanded_channels,
                kernel_size=cnf.kernel_size,
                stride=cnf.stride,
                padding=cnf.padding,
                groups=cnf.expanded_channels,
                bias=False,
                activation_layer=cnf.activation,
            )
        )

        if cnf.use_se is True:
            squeeze_channels = make_divisible(cnf.expanded_channels // 4, 8)
            layers.append(SqueezeExcitation(cnf.expanded_channels, squeeze_channels, scale_activation=nn.Hardsigmoid))

        # Project
        layers.append(
            Conv2dNormActivation(
                cnf.expanded_channels,
                cnf.out_channels,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=None,
            )
        )

        self.block = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.shortcut is True:
            return x + self.block(x)

        return self.block(x)


# pylint: disable=invalid-name
class MobileNet_v3(DetectorBackbone):
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        alpha: float = self.config["alpha"]
        large: bool = self.config["large"]
        self.mlp_head = self.config.get("mlp_head", True)

        if large is True:
            last_channels = int(round(1280 * max(1.0, alpha)))
            net_settings = [
                InvertedResidualConfig(16, 16, 16, (3, 3), (1, 1), (1, 1), alpha, False, nn.ReLU),
                InvertedResidualConfig(16, 24, 64, (3, 3), (2, 2), (1, 1), alpha, False, nn.ReLU),
                InvertedResidualConfig(24, 24, 72, (3, 3), (1, 1), (1, 1), alpha, False, nn.ReLU),
                InvertedResidualConfig(24, 40, 72, (5, 5), (2, 2), (2, 2), alpha, True, nn.ReLU),
                InvertedResidualConfig(40, 40, 120, (5, 5), (1, 1), (2, 2), alpha, True, nn.ReLU),
                InvertedResidualConfig(40, 40, 120, (5, 5), (1, 1), (2, 2), alpha, True, nn.ReLU),
                InvertedResidualConfig(40, 80, 240, (3, 3), (2, 2), (1, 1), alpha, False, nn.Hardswish),
                InvertedResidualConfig(80, 80, 200, (3, 3), (1, 1), (1, 1), alpha, False, nn.Hardswish),
                InvertedResidualConfig(80, 80, 184, (3, 3), (1, 1), (1, 1), alpha, False, nn.Hardswish),
                InvertedResidualConfig(80, 80, 184, (3, 3), (1, 1), (1, 1), alpha, False, nn.Hardswish),
                InvertedResidualConfig(80, 112, 480, (3, 3), (1, 1), (1, 1), alpha, True, nn.Hardswish),
                InvertedResidualConfig(112, 112, 672, (3, 3), (1, 1), (1, 1), alpha, True, nn.Hardswish),
                InvertedResidualConfig(112, 160, 672, (5, 5), (2, 2), (2, 2), alpha, True, nn.Hardswish),
                InvertedResidualConfig(160, 160, 960, (5, 5), (1, 1), (2, 2), alpha, True, nn.Hardswish),
                InvertedResidualConfig(160, 160, 960, (5, 5), (1, 1), (2, 2), alpha, True, nn.Hardswish),
            ]
        else:
            last_channels = int(round(1024 * max(1.0, alpha)))
            net_settings = [
                InvertedResidualConfig(16, 16, 16, (3, 3), (2, 2), (1, 1), alpha, True, nn.ReLU),
                InvertedResidualConfig(16, 24, 72, (3, 3), (2, 2), (1, 1), alpha, False, nn.ReLU),
                InvertedResidualConfig(24, 24, 88, (3, 3), (1, 1), (1, 1), alpha, False, nn.ReLU),
                InvertedResidualConfig(24, 40, 96, (5, 5), (2, 2), (2, 2), alpha, True, nn.Hardswish),
                InvertedResidualConfig(40, 40, 240, (5, 5), (1, 1), (2, 2), alpha, True, nn.Hardswish),
                InvertedResidualConfig(40, 40, 240, (5, 5), (1, 1), (2, 2), alpha, True, nn.Hardswish),
                InvertedResidualConfig(40, 48, 120, (5, 5), (1, 1), (2, 2), alpha, True, nn.Hardswish),
                InvertedResidualConfig(48, 48, 144, (5, 5), (1, 1), (2, 2), alpha, True, nn.Hardswish),
                InvertedResidualConfig(48, 96, 288, (5, 5), (2, 2), (2, 2), alpha, True, nn.Hardswish),
                InvertedResidualConfig(96, 96, 576, (5, 5), (1, 1), (2, 2), alpha, True, nn.Hardswish),
                InvertedResidualConfig(96, 96, 576, (5, 5), (1, 1), (2, 2), alpha, True, nn.Hardswish),
            ]

        self.stem = Conv2dNormActivation(
            self.input_channels,
            net_settings[0].in_channels,
            kernel_size=(3, 3),
            stride=(2, 2),
            padding=(1, 1),
            bias=False,
            activation_layer=nn.Hardswish,
        )

        layers: list[nn.Module] = []
        i = 0
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for idx, block_settings in enumerate(net_settings):
            if block_settings.stride[0] > 1 or block_settings.stride[1] > 1:
                stages[f"stage{i}"] = nn.Sequential(*layers)
                return_channels.append(net_settings[idx - 1].out_channels)
                layers = []
                i += 1

            layers.append(InvertedResidual(block_settings))

        stages[f"stage{i}"] = nn.Sequential(*layers)
        return_channels.append(net_settings[-1].out_channels)

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            Conv2dNormActivation(
                net_settings[-1].out_channels,
                net_settings[-1].out_channels * 6,
                kernel_size=(1, 1),
                stride=(1, 1),
                padding=(0, 0),
                bias=False,
                activation_layer=nn.Hardswish,
            ),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels[1:5]
        self.embedding_size = net_settings[-1].out_channels * 6
        self.last_channels = last_channels
        self.classifier = self.create_classifier()

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad_(False)

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad_(False)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def create_classifier(
        self, embed_dim: Optional[int] = None, head_bias: Optional[bool] = None, mlp_head: Optional[bool] = None
    ) -> nn.Module:
        if self.num_classes == 0:
            return nn.Identity()

        if embed_dim is None:
            embed_dim = self.embedding_size
        if head_bias is None:
            head_bias = self.head_bias
        if mlp_head is None:
            mlp_head = self.mlp_head

        if mlp_head is True:
            return nn.Sequential(
                nn.Linear(embed_dim, self.last_channels),
                nn.Hardswish(inplace=True),
                nn.Dropout(p=0.2, inplace=True),
                nn.Linear(self.last_channels, self.num_classes, bias=head_bias),
            )

        return nn.Linear(embed_dim, self.num_classes, bias=head_bias)


registry.register_model_config("mobilenet_v3_small_0_25", MobileNet_v3, config={"alpha": 0.25, "large": False})
registry.register_model_config("mobilenet_v3_small_0_5", MobileNet_v3, config={"alpha": 0.5, "large": False})
registry.register_model_config("mobilenet_v3_small_0_75", MobileNet_v3, config={"alpha": 0.75, "large": False})
registry.register_model_config("mobilenet_v3_small_1_0", MobileNet_v3, config={"alpha": 1.0, "large": False})
registry.register_model_config("mobilenet_v3_small_1_25", MobileNet_v3, config={"alpha": 1.25, "large": False})
registry.register_model_config("mobilenet_v3_small_1_5", MobileNet_v3, config={"alpha": 1.5, "large": False})
registry.register_model_config("mobilenet_v3_small_1_75", MobileNet_v3, config={"alpha": 1.75, "large": False})
registry.register_model_config("mobilenet_v3_small_2_0", MobileNet_v3, config={"alpha": 2.0, "large": False})

registry.register_model_config("mobilenet_v3_large_0_25", MobileNet_v3, config={"alpha": 0.25, "large": True})
registry.register_model_config("mobilenet_v3_large_0_5", MobileNet_v3, config={"alpha": 0.5, "large": True})
registry.register_model_config("mobilenet_v3_large_0_75", MobileNet_v3, config={"alpha": 0.75, "large": True})
registry.register_model_config("mobilenet_v3_large_1_0", MobileNet_v3, config={"alpha": 1.0, "large": True})
registry.register_model_config("mobilenet_v3_large_1_25", MobileNet_v3, config={"alpha": 1.25, "large": True})
registry.register_model_config("mobilenet_v3_large_1_5", MobileNet_v3, config={"alpha": 1.5, "large": True})
registry.register_model_config("mobilenet_v3_large_1_75", MobileNet_v3, config={"alpha": 1.75, "large": True})
registry.register_model_config("mobilenet_v3_large_2_0", MobileNet_v3, config={"alpha": 2.0, "large": True})


registry.register_weights(
    "mobilenet_v3_small_1_0_il-common",
    {
        "description": "MobileNet v3 small (1.0 multiplier) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 7.4,
                "sha256": "ac53227f7513fd0c0b5204ee57403de2ab6c74c4e4d1061b9168596c6b5cea48",
            }
        },
        "net": {"network": "mobilenet_v3_small_1_0", "tag": "il-common"},
    },
)
registry.register_weights(
    "mobilenet_v3_large_0_75_il-common",
    {
        "description": "MobileNet v3 large (0.75 multiplier) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 12.1,
                "sha256": "92412316f3dbcc41e4f3186acb50027e87ce0ea3ea1d6b5a726ea883fea20b8e",
            }
        },
        "net": {"network": "mobilenet_v3_large_0_75", "tag": "il-common"},
    },
)
registry.register_weights(
    "mobilenet_v3_large_1_0_il-common",
    {
        "description": "MobileNet v3 large (1.0 multiplier) model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 18.1,
                "sha256": "85bc65482716368c4fd69d4c699b61330fdf806d99e6c88ffbeda54886c2f3d4",
            }
        },
        "net": {"network": "mobilenet_v3_large_1_0", "tag": "il-common"},
    },
)



"""
Weighted Boxes Fusion, adapted from
https://github.com/ZFTurbo/Weighted-Boxes-Fusion

Paper "Weighted boxes fusion: Ensembling boxes from different object detection models",
https://arxiv.org/abs/1910.13302
"""

# Reference license: MIT

from dataclasses import dataclass
from typing import Literal
from typing import Optional

import torch
from torchvision.ops import box_iou

ConfType = Literal["avg", "max", "box_and_model_avg", "absent_model_aware_avg"]


@dataclass
class BoxCluster:
    box: torch.Tensor
    score_weight_sum: torch.Tensor
    weight_sum: torch.Tensor
    max_score: torch.Tensor
    boxes_count: int

    @classmethod
    def from_entry(cls, box: torch.Tensor, score: torch.Tensor, weight: torch.Tensor) -> "BoxCluster":
        score_weight = score * weight
        return cls(
            box=box.clone(),
            score_weight_sum=score_weight,
            weight_sum=weight,
            max_score=score,
            boxes_count=1,
        )

    def add(self, box: torch.Tensor, score: torch.Tensor, weight: torch.Tensor) -> None:
        score_weight = score * weight
        total_weight = self.score_weight_sum + score_weight
        self.box = (self.box * self.score_weight_sum + box * score_weight) / total_weight
        self.score_weight_sum = total_weight
        self.weight_sum += weight
        self.max_score = torch.maximum(self.max_score, score)
        self.boxes_count += 1


# pylint: disable=too-many-locals,too-many-branches
def weighted_boxes_fusion(
    boxes_list: list[torch.Tensor],
    scores_list: list[torch.Tensor],
    labels_list: list[torch.Tensor],
    weights: Optional[list[float]] = None,
    iou_thr: float = 0.55,
    skip_box_thr: float = 0.0,
    conf_type: ConfType = "avg",
    allows_overflow: bool = False,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    if weights is None:
        weights = [1.0] * len(boxes_list)
    if len(weights) != len(boxes_list):
        raise ValueError("weights must match number of box sets")

    if len(boxes_list) > 0:
        device = boxes_list[0].device
    else:
        device = torch.device("cpu")

    boxes_all: list[torch.Tensor] = []
    scores_all: list[torch.Tensor] = []
    labels_all: list[torch.Tensor] = []
    weights_all: list[torch.Tensor] = []
    for boxes, scores, labels, weight in zip(boxes_list, scores_list, labels_list, weights):
        if boxes.numel() == 0:
            continue

        boxes_tensor = boxes.detach().to(dtype=torch.float32)
        scores_tensor = scores.detach().to(dtype=torch.float32)
        labels_tensor = labels.detach().to(dtype=torch.int64)

        keep = scores_tensor >= skip_box_thr
        if not keep.any():
            continue

        boxes_tensor = boxes_tensor[keep]
        scores_tensor = scores_tensor[keep]
        labels_tensor = labels_tensor[keep]
        weights_tensor = scores_tensor.new_full(scores_tensor.shape, weight)

        boxes_all.append(boxes_tensor)
        scores_all.append(scores_tensor)
        labels_all.append(labels_tensor)
        weights_all.append(weights_tensor)

    if len(boxes_all) == 0:
        empty_boxes = torch.zeros((0, 4), dtype=torch.float32, device=device)
        empty_scores = torch.zeros((0,), dtype=torch.float32, device=device)
        empty_labels = torch.zeros((0,), dtype=torch.int64, device=device)
        return (empty_boxes, empty_scores, empty_labels)

    boxes_tensor = torch.concat(boxes_all, dim=0)
    scores_tensor = torch.concat(scores_all, dim=0)
    labels_tensor = torch.concat(labels_all, dim=0)
    weights_tensor = torch.concat(weights_all, dim=0)
    labels_unique = torch.unique(labels_tensor)

    total_weight = float(sum(weights))
    num_models = len(weights)
    fused_boxes: list[torch.Tensor] = []
    fused_scores: list[torch.Tensor] = []
    fused_labels: list[torch.Tensor] = []

    for label in labels_unique:
        label_mask = labels_tensor == label
        label_boxes = boxes_tensor[label_mask]
        label_scores = scores_tensor[label_mask]
        label_weights = weights_tensor[label_mask]
        order = torch.argsort(label_scores, descending=True)
        clusters: list[BoxCluster] = []
        for idx in order:
            box = label_boxes[idx]
            score = label_scores[idx]
            weight = label_weights[idx]
            if len(clusters) == 0:
                clusters.append(BoxCluster.from_entry(box, score, weight))
                continue

            cluster_boxes = torch.stack([cluster.box for cluster in clusters], dim=0)
            ious = box_iou(box.unsqueeze(0), cluster_boxes).squeeze(0)
            max_iou, max_idx = torch.max(ious, dim=0)
            if max_iou > iou_thr:
                clusters[int(max_idx)].add(box, score, weight)
            else:
                clusters.append(BoxCluster.from_entry(box, score, weight))

        for cluster in clusters:
            if conf_type == "avg":
                score = cluster.score_weight_sum / cluster.weight_sum
            elif conf_type == "max":
                score = cluster.max_score
            elif conf_type == "box_and_model_avg":
                score = (cluster.score_weight_sum / cluster.weight_sum) * (cluster.boxes_count / num_models)
            elif conf_type == "absent_model_aware_avg":
                score = cluster.score_weight_sum / total_weight
            else:
                raise ValueError(f"Unsupported conf_type: {conf_type}")

            if allows_overflow is False:
                score = score.clamp(max=1.0)

            fused_boxes.append(cluster.box)
            fused_scores.append(score)
            fused_labels.append(label)

    fused_scores_tensor = torch.stack(fused_scores)
    order = torch.argsort(fused_scores_tensor, descending=True)
    fused_boxes_tensor = torch.stack(fused_boxes, dim=0)[order]
    fused_scores_tensor = fused_scores_tensor[order]
    fused_labels_tensor = torch.stack(fused_labels)[order]

    return (fused_boxes_tensor, fused_scores_tensor, fused_labels_tensor)


def fuse_detections_wbf_single(
    detections: list[dict[str, torch.Tensor]],
    weights: Optional[list[float]] = None,
    iou_thr: float = 0.55,
    skip_box_thr: float = 0.0,
    conf_type: ConfType = "avg",
    allows_overflow: bool = False,
) -> dict[str, torch.Tensor]:
    if len(detections) == 0:
        return {
            "boxes": torch.zeros((0, 4)),
            "scores": torch.zeros((0,)),
            "labels": torch.zeros((0,), dtype=torch.int64),
        }

    boxes_list = [detection["boxes"] for detection in detections]
    scores_list = [detection["scores"] for detection in detections]
    labels_list = [detection["labels"] for detection in detections]

    boxes, scores, labels = weighted_boxes_fusion(
        boxes_list,
        scores_list,
        labels_list,
        weights=weights,
        iou_thr=iou_thr,
        skip_box_thr=skip_box_thr,
        conf_type=conf_type,
        allows_overflow=allows_overflow,
    )

    return {"boxes": boxes, "scores": scores, "labels": labels}


def fuse_detections_wbf(
    detections_list: list[list[dict[str, torch.Tensor]]],
    weights: Optional[list[float]] = None,
    iou_thr: float = 0.55,
    skip_box_thr: float = 0.0,
    conf_type: ConfType = "avg",
    allows_overflow: bool = False,
) -> list[dict[str, torch.Tensor]]:
    if len(detections_list) == 0:
        return []

    # Outer list is the augmentations, inner is the batch
    batch_size = len(detections_list[0])
    fused: list[dict[str, torch.Tensor]] = []
    for idx in range(batch_size):
        per_image = [detections[idx] for detections in detections_list]
        fused.append(
            fuse_detections_wbf_single(
                per_image,
                weights=weights,
                iou_thr=iou_thr,
                skip_box_thr=skip_box_thr,
                conf_type=conf_type,
                allows_overflow=allows_overflow,
            )
        )

    return fused



"""
Paper "SimpleShot: Revisiting Nearest-Neighbor Classification for Few-Shot Learning",
https://arxiv.org/abs/1911.04623
"""

import numpy as np
import numpy.typing as npt
from sklearn.neighbors import NearestCentroid

from birder.eval._embeddings import l2_normalize


def sample_k_shot(
    features: npt.NDArray[np.float32], labels: npt.NDArray[np.int_], k: int, rng: np.random.Generator
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.int_]]:
    """
    Sample k examples per class from the training set

    Returns
    -------
    sampled_features
        Features of shape (num_classes * k, embedding_dim).
    sampled_labels
        Labels of shape (num_classes * k,).
    """

    unique_labels = np.unique(labels)
    sampled_features_list: list[npt.NDArray[np.float32]] = []
    sampled_labels_list: list[npt.NDArray[np.int_]] = []

    for label in unique_labels:
        mask = labels == label
        class_features = features[mask]
        class_labels = labels[mask]

        # Sample k examples (with replacement if fewer than k available)
        n_available = len(class_features)
        if n_available >= k:
            indices = rng.choice(n_available, size=k, replace=False)
        else:
            indices = rng.choice(n_available, size=k, replace=True)

        sampled_features_list.append(class_features[indices])
        sampled_labels_list.append(class_labels[indices])

    return (np.concatenate(sampled_features_list), np.concatenate(sampled_labels_list))


def normalize_features(
    train_features: npt.NDArray[np.float32], test_features: npt.NDArray[np.float32]
) -> tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]:
    """
    Apply SimpleShot normalization: mean centering followed by L2 normalization
    """

    mean = train_features.mean(axis=0, keepdims=True)
    train_centered = train_features - mean
    test_centered = test_features - mean

    train_norm = l2_normalize(train_centered)
    test_norm = l2_normalize(test_centered)

    return (train_norm, test_norm)


def evaluate_simpleshot(
    train_features: npt.NDArray[np.float32],
    train_labels: npt.NDArray[np.int_],
    test_features: npt.NDArray[np.float32],
    test_labels: npt.NDArray[np.int_],
) -> tuple[npt.NDArray[np.int_], npt.NDArray[np.int_]]:
    """
    Evaluate using SimpleShot method

    Parameters
    ----------
    train_features
        Training features of shape (n_train, embedding_dim).
    train_labels
        Training labels of shape (n_train,).
    test_features
        Test features of shape (n_test, embedding_dim).
    test_labels
        Test labels of shape (n_test,).

    Returns
    -------
    y_pred
        Predicted labels for test samples.
    y_true
        True labels for test samples (same as test_labels).
    """

    train_norm, test_norm = normalize_features(train_features, test_features)

    clf = NearestCentroid()
    clf.fit(train_norm, train_labels)
    y_pred = clf.predict(test_norm)

    return (y_pred, test_labels)



"""
LIT v1 Tiny, adapted from
https://github.com/ziplab/LIT/blob/main/classification/code_for_lit_ti/lit.py

Paper "Less is More: Pay Less Attention in Vision Transformers", https://arxiv.org/abs/2105.14217

Generated by Claude Code Opus 4.5
"""

# Reference license: Apache-2.0

import math
from collections import OrderedDict
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.lit_v1 import MLP
from birder.net.lit_v1 import DeformablePatchMerging
from birder.net.lit_v1 import IdentityDownsample
from birder.net.vit import adjust_position_embedding


class MLPBlock(nn.Module):
    def __init__(self, dim: int, mlp_ratio: float, drop_path: float) -> None:
        super().__init__()
        self.norm = nn.LayerNorm(dim, eps=1e-6)
        self.mlp = MLP(dim, int(dim * mlp_ratio))
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x + self.drop_path(self.mlp(self.norm(x)))


class Attention(nn.Module):
    def __init__(self, dim: int, num_heads: int) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.scale = (dim // num_heads) ** -0.5
        self.qkv = nn.Linear(dim, dim * 3)
        self.proj = nn.Linear(dim, dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        B, N, C = x.size()
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = F.softmax(attn, dim=-1)

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)

        return x


class ViTBlock(nn.Module):
    def __init__(self, dim: int, num_heads: int, mlp_ratio: float, drop_path: float) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim, eps=1e-6)
        self.attn = Attention(dim, num_heads)
        self.drop_path1 = StochasticDepth(drop_path, mode="row")

        self.norm2 = nn.LayerNorm(dim, eps=1e-6)
        self.mlp = MLP(dim, int(dim * mlp_ratio))
        self.drop_path2 = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.drop_path1(self.attn(self.norm1(x)))
        x = x + self.drop_path2(self.mlp(self.norm2(x)))

        return x


class LITStage(nn.Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        input_resolution: tuple[int, int],
        depth: int,
        num_heads: int,
        mlp_ratio: float,
        has_msa: bool,
        downsample: bool,
        use_cls_token: bool,
        drop_path: list[float],
    ) -> None:
        super().__init__()
        self.dynamic_size = False
        self.input_resolution = input_resolution
        self.downsample: nn.Module
        if downsample is True:
            self.downsample = DeformablePatchMerging(in_dim, out_dim)
        else:
            self.downsample = IdentityDownsample()

        blocks: list[nn.Module] = []
        for i in range(depth):
            if has_msa is True:
                blocks.append(ViTBlock(out_dim, num_heads, mlp_ratio, drop_path[i]))
            else:
                blocks.append(MLPBlock(out_dim, mlp_ratio, drop_path[i]))

        self.blocks = nn.ModuleList(blocks)

        num_tokens = input_resolution[0] * input_resolution[1]
        if use_cls_token is True:
            self.cls_token = nn.Parameter(torch.zeros(1, 1, out_dim))
            nn.init.trunc_normal_(self.cls_token, std=0.02)
            num_tokens += 1
        else:
            self.cls_token = None

        self.pos_embed = nn.Parameter(torch.zeros(1, num_tokens, out_dim))
        nn.init.trunc_normal_(self.pos_embed, std=0.02)

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        self.dynamic_size = dynamic_size

    def _get_pos_embed(self, H: int, W: int) -> torch.Tensor:
        if self.dynamic_size is False or (H == self.input_resolution[0] and W == self.input_resolution[1]):
            return self.pos_embed

        if self.cls_token is not None:
            num_prefix_tokens = 1
        else:
            num_prefix_tokens = 0

        return adjust_position_embedding(
            self.pos_embed, self.input_resolution, (H, W), num_prefix_tokens=num_prefix_tokens
        )

    def forward(self, x: torch.Tensor, input_resolution: tuple[int, int]) -> tuple[torch.Tensor, int, int]:
        x, H, W = self.downsample(x, input_resolution)

        if self.cls_token is not None:
            cls_tokens = self.cls_token.expand(x.size(0), -1, -1)
            x = torch.concat((cls_tokens, x), dim=1)

        x = x + self._get_pos_embed(H, W)

        for block in self.blocks:
            x = block(x)

        return (x, H, W)


# pylint: disable=invalid-name
class LIT_v1_Tiny(DetectorBackbone):
    block_group_regex = r"body\.stage(\d+)\.blocks\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        patch_size = 4
        stage_dims: list[int] = self.config["stage_dims"]
        depths: list[int] = self.config["depths"]
        num_heads: list[int] = self.config["num_heads"]
        mlp_ratios: list[float] = self.config["mlp_ratios"]
        has_msa: list[bool] = self.config["has_msa"]
        drop_path_rate: float = self.config["drop_path_rate"]

        num_stages = len(depths)

        self.stem = nn.Sequential(
            nn.Conv2d(
                self.input_channels,
                stage_dims[0],
                kernel_size=(patch_size, patch_size),
                stride=(patch_size, patch_size),
                padding=(0, 0),
            ),
            Permute([0, 2, 3, 1]),
            nn.LayerNorm(stage_dims[0], eps=1e-6),
        )

        # Stochastic depth
        dpr = [x.tolist() for x in torch.linspace(0, drop_path_rate, sum(depths)).split(depths)]

        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        resolution = (self.size[0] // patch_size, self.size[1] // patch_size)

        for i in range(num_stages):
            if i > 0:
                resolution = (resolution[0] // 2, resolution[1] // 2)

            stage = LITStage(
                stage_dims[i - 1] if i > 0 else stage_dims[0],
                stage_dims[i],
                input_resolution=resolution,
                depth=depths[i],
                num_heads=num_heads[i],
                mlp_ratio=mlp_ratios[i],
                has_msa=has_msa[i],
                downsample=i > 0,
                use_cls_token=i == num_stages - 1,
                drop_path=dpr[i],
            )
            stages[f"stage{i + 1}"] = stage
            return_channels.append(stage_dims[i])

        self.body = nn.ModuleDict(stages)
        self.norm = nn.LayerNorm(stage_dims[-1], eps=1e-6)
        self.return_channels = return_channels
        self.embedding_size = stage_dims[-1]
        self.classifier = self.create_classifier()
        self.patch_size = patch_size

        # Weight initialization
        for name, m in self.named_modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Conv2d):
                if name.endswith("offset_conv") is True:
                    continue

                fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                fan_out //= m.groups
                nn.init.normal_(m.weight, mean=0.0, std=math.sqrt(2.0 / fan_out))
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)
        B, H, W, C = x.size()
        x = x.reshape(B, H * W, C)

        out = {}
        for name, stage in self.body.items():
            x, H, W = stage(x, (H, W))
            if name in self.return_stages:
                if stage.cls_token is not None:
                    spatial_x = x[:, 1:]
                else:
                    spatial_x = x

                out[name] = spatial_x.reshape(B, H, W, -1).permute(0, 3, 1, 2).contiguous()

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad_(False)

        for idx, stage in enumerate(self.body.values()):
            if idx >= up_to_stage:
                break

            for param in stage.parameters():
                param.requires_grad_(False)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        B, H, W, C = x.size()
        x = x.reshape(B, H * W, C)
        for stage in self.body.values():
            x, H, W = stage(x, (H, W))

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        x = self.norm(x)
        return x[:, 0]

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        super().set_dynamic_size(dynamic_size)
        for stage in self.body.values():
            stage.set_dynamic_size(dynamic_size)

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)

        new_patches_resolution = (new_size[0] // self.patch_size, new_size[1] // self.patch_size)

        h, w = new_patches_resolution
        for stage in self.body.values():
            if not isinstance(stage.downsample, IdentityDownsample):
                h = h // 2
                w = w // 2

            out_resolution = (h, w)
            if out_resolution == stage.input_resolution:
                continue

            if stage.cls_token is not None:
                num_prefix_tokens = 1
            else:
                num_prefix_tokens = 0

            with torch.no_grad():
                pos_embed = adjust_position_embedding(
                    stage.pos_embed,
                    stage.input_resolution,
                    out_resolution,
                    num_prefix_tokens=num_prefix_tokens,
                )

            stage.input_resolution = out_resolution
            stage.pos_embed = nn.Parameter(pos_embed)


registry.register_model_config(
    "lit_v1_t",
    LIT_v1_Tiny,
    config={
        "stage_dims": [64, 128, 320, 512],
        "depths": [3, 4, 6, 3],
        "num_heads": [1, 2, 5, 8],
        "mlp_ratios": [8.0, 8.0, 4.0, 4.0],
        "has_msa": [False, False, True, True],
        "drop_path_rate": 0.1,
    },
)

registry.register_weights(
    "lit_v1_t_il-common",
    {
        "description": "LIT v1 Tiny model trained on the il-common dataset",
        "resolution": (256, 256),
        "formats": {
            "pt": {
                "file_size": 75.2,
                "sha256": "93813b2716eb9f33e06dc15ab2ba335c6d219354d2983bbc4f834f8f4e688e5c",
            }
        },
        "net": {"network": "lit_v1_t", "tag": "il-common"},
    },
)



"""
ConvNeXt v1 Isotropic, adapted from
https://github.com/facebookresearch/ConvNeXt/blob/main/models/convnext_isotropic.py

Paper "A ConvNet for the 2020s", https://arxiv.org/abs/2201.03545
"""

# Reference license: MIT

from functools import partial
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.layers import LayerNorm2d
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.base import normalize_out_indices


class ConvNeXtBlock(nn.Module):
    def __init__(self, channels: int, stochastic_depth_prob: float) -> None:
        super().__init__()
        self.block = nn.Sequential(
            nn.Conv2d(channels, channels, kernel_size=(7, 7), stride=(1, 1), padding=(3, 3), groups=channels),
            Permute([0, 2, 3, 1]),
            nn.LayerNorm(channels, eps=1e-6),
            nn.Linear(channels, 4 * channels),
            nn.GELU(),
            nn.Linear(4 * channels, channels),
            Permute([0, 3, 1, 2]),
        )
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        x = self.block(x)
        x = self.stochastic_depth(x)
        x += identity

        return x


# pylint: disable=invalid-name
class ConvNeXt_v1_Isotropic(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.(\d+)"

    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        patch_size = 16
        dim: int = self.config["dim"]
        num_layers: int = self.config["num_layers"]
        out_indices: Optional[list[int]] = self.config.get("out_indices", None)
        drop_path_rate: float = self.config["drop_path_rate"]

        torch._assert(self.size[0] % patch_size == 0, "Input shape indivisible by patch size!")
        torch._assert(self.size[1] % patch_size == 0, "Input shape indivisible by patch size!")
        self.patch_size = patch_size
        self.out_indices = normalize_out_indices(out_indices, num_layers)

        self.stem = nn.Conv2d(
            self.input_channels,
            dim,
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            padding=(0, 0),
        )

        layers = []
        for idx in range(num_layers):
            # Adjust stochastic depth probability based on the depth of the stage block
            sd_prob = drop_path_rate * idx / (num_layers - 1.0)
            layers.append(ConvNeXtBlock(dim, sd_prob))

        self.body = nn.Sequential(*layers)
        self.features = nn.Sequential(
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            LayerNorm2d(dim, eps=1e-6),
            nn.Flatten(1),
        )

        num_return_stages = len(self.out_indices) if self.out_indices is not None else 1
        self.return_stages = [f"stage{stage_idx + 1}" for stage_idx in range(num_return_stages)]
        self.return_channels = [dim] * num_return_stages
        self.embedding_size = dim
        self.classifier = self.create_classifier()

        self.max_stride = patch_size
        self.stem_stride = patch_size
        self.stem_width = dim
        self.encoding_size = dim
        self.decoder_block = partial(ConvNeXtBlock, stochastic_depth_prob=0)

        # Weights initialization
        for m in self.modules():
            if isinstance(m, (nn.Conv2d, nn.Linear)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        if self.out_indices is None:
            x = self.body(x)
            return {self.return_stages[0]: x}

        stage_num = 0
        out: dict[str, torch.Tensor] = {}
        for idx, module in enumerate(self.body.children()):
            x = module(x)
            if idx in self.out_indices:
                out[self.return_stages[stage_num]] = x
                stage_num += 1

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad_(False)

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad_(False)

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        assert new_size[0] % self.patch_size == 0, "Input shape indivisible by patch size!"
        assert new_size[1] % self.patch_size == 0, "Input shape indivisible by patch size!"

        super().adjust_size(new_size)


registry.register_model_config(
    "convnext_v1_iso_small",
    ConvNeXt_v1_Isotropic,
    config={"dim": 384, "num_layers": 18, "drop_path_rate": 0.1},
)
registry.register_model_config(
    "convnext_v1_iso_base",
    ConvNeXt_v1_Isotropic,
    config={"in_channels": 768, "num_layers": 18, "drop_path_rate": 0.2},
)
registry.register_model_config(
    "convnext_v1_iso_large",
    ConvNeXt_v1_Isotropic,
    config={"in_channels": 1024, "num_layers": 36, "drop_path_rate": 0.5},
)



"""
Hiera, adapted from
https://github.com/facebookresearch/hiera/blob/main/hiera/hiera.py

Paper "Hiera: A Hierarchical Vision Transformer without the Bells-and-Whistles", https://arxiv.org/abs/2306.00989

AbsWin variant from the paper "Window Attention is Bugged: How not to Interpolate Position Embeddings",
https://arxiv.org/abs/2311.05613

Changes from original:
* Support only 2d
"""

# Reference license: Apache-2.0

import math
from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_from_indices
from birder.layers import MultiHeadAttentionPool
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenOmissionResultType
from birder.net.vit import adjust_position_embedding


def get_resized_mask(target_size: list[int], mask: torch.Tensor) -> torch.Tensor:
    # target_size: [H, W]
    # mask: [B, C, H, W]
    if mask is None:
        return mask

    assert len(mask.shape[2:]) == len(target_size)
    if mask.shape[2:] != target_size:
        return F.interpolate(mask.float(), size=target_size)

    return mask


def undo_windowing(x: torch.Tensor, shape: tuple[int, int]) -> torch.Tensor:
    """
    Restore spatial organization by undoing windowed organization of mask units

    Parameters
    ----------
    x
        Organized by mask units windows, e.g. in 2d [B, #mu_y*#mu_x, mu_y, mu_x, C]
    shape
        Current spatial shape, if it were not organized into mask unit windows,
        e.g. in 2d [B, #mu_y*mu_y, #mu_x*mu_x, C]

    Returns
    -------
    Tensor with shape of [B, #mu_y*mu_y, #mu_x*mu_x, C]
    """

    # [B, #mu_y*#mu_x, mu_y, mu_x, C] -> [B, #mu_y, #mu_x, mu_y, mu_x, C]
    B, _, mu_y, mu_x, C = x.size()
    H, W = shape
    nwy = H // mu_y
    nwx = W // mu_x
    x = x.view(B, nwy, nwx, mu_y, mu_x, C)

    # [B, #mu_y, #mu_x, mu_y, mu_x, C] -> [B, #mu_y*mu_y, #mu_x*mu_x, C]
    x = x.permute(0, 1, 3, 2, 4, 5).reshape(B, H, W, C)

    return x


class AvgTokens(nn.Module):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.mean(dim=1)


class PatchEmbed(nn.Module):
    def __init__(
        self, dim_in: int, dim_out: int, kernel: tuple[int, int], stride: tuple[int, int], padding: tuple[int, int]
    ) -> None:
        super().__init__()
        self.proj = nn.Conv2d(dim_in, dim_out, kernel_size=kernel, stride=stride, padding=padding)

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        if mask is not None:
            mask = get_resized_mask(target_size=x.shape[2:], mask=mask)
            x = self.proj(x * mask.to(torch.bool))
        else:
            x = self.proj(x)

        x = x.reshape(x.shape[0], x.shape[1], -1).transpose(2, 1)

        return x


class Unroll(nn.Module):
    """
    Reorders the tokens such that patches are contiguous in memory.
    E.g., given [B, (H, W), C] and stride of (Sy, Sx), this will re-order the tokens as
                           [B, (Sy, Sx, H // Sy, W // Sx), C]

    This allows operations like Max2d to be computed as x.view(B, Sx*Sy, -1, C).max(dim=1).
    Not only is this faster, but it also makes it easy to support inputs of arbitrary
    dimensions in addition to patch-wise sparsity.

    Performing this operation multiple times in sequence puts entire windows as contiguous
    in memory. For instance, if you applied the stride (2, 2) 3 times, entire windows of
    size 8x8 would be contiguous in memory, allowing operations like mask unit attention
    computed easily and efficiently, while also allowing max to be applied sequentially.

    Note: This means that intermediate values of the model are not in HxW order, so they
    need to be re-rolled if you want to use the intermediate values as a HxW feature map.
    The last block of the network is fine though, since by then the strides are all consumed.
    """

    def __init__(
        self, input_size: tuple[int, int], patch_stride: tuple[int, int], unroll_schedule: list[tuple[int, int]]
    ) -> None:
        super().__init__()
        self.size = (input_size[0] // patch_stride[0], input_size[1] // patch_stride[1])
        self.schedule = unroll_schedule

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        B, _, C = x.shape
        H, W = self.size
        x = x.view(B, H, W, C)

        for sy, sx in self.schedule:
            h2 = H // sy
            w2 = W // sx
            # [B, h2, sy, w2, sx, C] -> [B, sy, sx, h2, w2, C]
            x = x.view(B, h2, sy, w2, sx, C).permute(0, 2, 4, 1, 3, 5)

            # Flatten sy, sx into batch
            x = x.reshape(B * sy * sx, h2, w2, C)
            B = B * sy * sx
            H = h2
            W = w2

        x = x.reshape(-1, self.size[0] * self.size[1], C)

        return x


class Reroll(nn.Module):
    """
    Undo the "unroll" operation so that intermediate features are usable
    """

    def __init__(
        self,
        input_size: tuple[int, int],
        patch_stride: tuple[int, int],
        unroll_schedule: list[tuple[int, int]],
        stage_ends: list[int],
        q_pool: int,
    ):
        super().__init__()
        self.size = (input_size[0] // patch_stride[0], input_size[1] // patch_stride[1])

        # The first stage has to reverse everything
        # The next stage has to reverse all but the first unroll, etc.
        self.schedule: dict[int, tuple[list[tuple[int, int]], tuple[int, int]]] = {}
        size = self.size
        for i in range(stage_ends[-1] + 1):
            self.schedule[i] = (unroll_schedule, size)
            # Schedule unchanged if no pooling at a stage end
            if i in stage_ends[:q_pool]:
                if len(unroll_schedule) > 0:
                    size = (size[0] // unroll_schedule[0][0], size[1] // unroll_schedule[0][1])

                unroll_schedule = unroll_schedule[1:]

    def forward(self, x: torch.Tensor, block_idx: int, mask: torch.Tensor = None) -> torch.Tensor:
        """
        Roll the given tensor back up to spatial order assuming it's from the given block.

        If no mask is provided:
            - Returns [B, H, W, C]
        If a mask is provided:
            - Returns [B, #MUs, mu_y, mu_x, C]
        """

        schedule, size = self.schedule[block_idx]
        B, N, C = x.size()

        mu_y = 1
        mu_x = 1
        for sy, sx in schedule:
            x = x.view(B, sy, sx, N // (sy * sx), mu_y, mu_x, C)

            # -> [B, N//..., Sy, mu_y, Sx, mu_x, C]
            x = x.permute(0, 3, 1, 4, 2, 5, 6)

            mu_y = mu_y * sy
            mu_x = mu_x * sx
            x = x.reshape(B, -1, mu_y, mu_x, C)
            N = x.size(1)

        x = x.reshape(B, N, mu_y, mu_x, C)

        # If masked, return [B, #MUs, mu_y, mu_x, C]
        if mask is not None:
            return x

        # If not masked, we can return [B, H, W, C]
        x = undo_windowing(x, size)

        return x


class MaskUnitAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        heads: int,
        q_stride: int,
        window_size: int,
        use_mask_unit_attn: bool,
    ) -> None:
        super().__init__()
        self.dim_out = dim_out
        self.heads = heads
        self.q_stride = q_stride
        self.head_dim = dim_out // heads
        self.scale = self.head_dim**-0.5
        self.window_size = window_size
        self.use_mask_unit_attn = use_mask_unit_attn

        self.qkv = nn.Linear(dim, 3 * dim_out)
        self.proj = nn.Linear(dim_out, dim_out)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        B, N, _ = x.shape
        if self.use_mask_unit_attn is True:
            num_windows = N // (self.q_stride * self.window_size)
        else:
            num_windows = 1

        qkv = self.qkv(x).reshape(B, -1, num_windows, 3, self.heads, self.head_dim).permute(3, 0, 4, 2, 1, 5)
        q, k, v = qkv.unbind(0)

        if self.q_stride > 1:
            # Refer to Unroll to see how this performs a maxpool-Nd
            q = q.view(B, self.heads, num_windows, self.q_stride, -1, self.head_dim).amax(dim=3)

        x = F.scaled_dot_product_attention(q, k, v, scale=self.scale)  # pylint: disable=not-callable

        x = x.transpose(1, 3).reshape(B, -1, self.dim_out)
        x = self.proj(x)

        return x


class HieraBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        dim_out: int,
        heads: int,
        mlp_ratio: float,
        drop_path: float,
        q_stride: int,
        window_size: int,
        use_mask_unit_attn: bool,
    ) -> None:
        super().__init__()
        self.dim = dim
        self.dim_out = dim_out

        self.norm1 = nn.LayerNorm(dim, eps=1e-6)
        if dim != dim_out:
            self.proj = nn.Linear(dim, dim_out)
        else:
            self.proj = None

        self.attn = MaskUnitAttention(dim, dim_out, heads, q_stride, window_size, use_mask_unit_attn)
        self.norm2 = nn.LayerNorm(dim_out, eps=1e-6)
        self.mlp = MLP(dim_out, [int(dim_out * mlp_ratio), dim_out], activation_layer=nn.GELU)
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_norm = self.norm1(x)
        if self.proj is not None:
            x = self.proj(x_norm)
            x = x.view(x.shape[0], self.attn.q_stride, -1, x.shape[-1]).amax(dim=1)  # Max-pool

        x = x + self.drop_path(self.attn(x_norm))
        x = x + self.drop_path(self.mlp(self.norm2(x)))

        return x


# pylint: disable=too-many-instance-attributes
class Hiera(DetectorBackbone, PreTrainEncoder, MaskedTokenOmissionMixin):
    scriptable = False
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        image_size = self.size
        mask_unit_size = (8, 8)
        q_stride = (2, 2)
        q_pool = 3
        patch_kernel = (7, 7)
        patch_stride = (4, 4)
        patch_padding = (3, 3)
        mask_unit_attn = (True, True, False, False)
        dim_mul = 2.0
        head_mul = 2.0
        mlp_ratio = 4.0
        depths: list[int] = self.config["depths"]
        embed_dim: int = self.config["embed_dim"]
        num_heads: int = self.config["num_heads"]
        abs_win_pos_embed: bool = self.config["abs_win_pos_embed"]
        attn_pool_head: bool = self.config.get("attn_pool_head", False)
        attn_pool_num_heads: Optional[int] = self.config.get("attn_pool_num_heads", None)
        drop_path_rate: float = self.config["drop_path_rate"]

        tokens_spatial_shape = [i // s for i, s in zip(image_size, patch_stride)]
        num_tokens = math.prod(tokens_spatial_shape)
        flat_mu_size = math.prod(mask_unit_size)
        flat_q_stride = math.prod(q_stride)
        assert q_pool < len(depths)
        assert patch_stride[0] == patch_stride[1]

        self.tokens_spatial_shape = tokens_spatial_shape
        self.mask_unit_size = mask_unit_size
        self.patch_stride = patch_stride
        self.q_stride = q_stride
        self.q_pool = q_pool
        self.mu_size = flat_mu_size
        self.mask_spatial_shape = [i // s for i, s in zip(tokens_spatial_shape, mask_unit_size)]
        self.stage_ends = [sum(depths[:i]) - 1 for i in range(1, len(depths) + 1)]
        self.num_special_tokens = 0
        self.num_layers = sum(depths)

        stem_dim = embed_dim
        self.stem = PatchEmbed(self.input_channels, stem_dim, patch_kernel, patch_stride, patch_padding)

        global_pos_size = (image_size[0] // 2**4, image_size[1] // 2**4)
        if abs_win_pos_embed is True:
            # AbsWin variant
            self.pos_embed = nn.Parameter(torch.zeros(1, embed_dim, *global_pos_size))
            self.pos_embed_win = nn.Parameter(torch.zeros(1, embed_dim, *mask_unit_size))
        else:
            self.pos_embed = nn.Parameter(torch.zeros(1, num_tokens, embed_dim))
            self.pos_embed_win = None

        self.unroll = Unroll(image_size, patch_stride, [q_stride] * len(self.stage_ends[:-1]))
        self.reroll = Reroll(image_size, patch_stride, [q_stride] * len(self.stage_ends[:-1]), self.stage_ends, q_pool)

        q_pool_blocks = [x + 1 for x in self.stage_ends[:q_pool]]
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, self.num_layers)]  # Stochastic depth decay rule

        cur_stage = 0
        layers = []
        self.block_count = []
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        for i in range(self.num_layers):
            dim_out = embed_dim

            # Mask unit or global attention.
            # Lag by 1 block, so that global attention, applied post pooling on lower resolution
            use_mask_unit_attn = mask_unit_attn[cur_stage]

            if i - 1 in self.stage_ends:
                dim_out = int(embed_dim * dim_mul)
                num_heads = int(num_heads * head_mul)
                cur_stage += 1
                if i in q_pool_blocks:
                    flat_mu_size //= flat_q_stride

            layers.append(
                HieraBlock(
                    dim=embed_dim,
                    dim_out=dim_out,
                    heads=num_heads,
                    mlp_ratio=mlp_ratio,
                    drop_path=dpr[i],
                    q_stride=(flat_q_stride if i in q_pool_blocks else 1),
                    window_size=flat_mu_size,
                    use_mask_unit_attn=use_mask_unit_attn,
                )
            )
            embed_dim = dim_out

            if i in self.stage_ends:
                stages[f"stage{cur_stage+1}"] = nn.Sequential(*layers)
                return_channels.append(dim_out)
                self.block_count.append(i)
                layers = []

        assert len(layers) == 0

        if attn_pool_head is False:
            attn_pool = None
        else:
            if attn_pool_num_heads is None:
                attn_pool_num_heads = num_heads

            attn_pool = MultiHeadAttentionPool(
                embed_dim, attn_pool_num_heads, int(mlp_ratio * embed_dim), qkv_bias=True
            )

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            attn_pool if attn_pool is not None else AvgTokens(),
            nn.LayerNorm(embed_dim, eps=1e-6),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = embed_dim
        self.classifier = self.create_classifier()

        self.max_stride = patch_stride[0] * 8
        self.stem_stride = patch_stride[0]
        self.stem_width = stem_dim
        self.encoding_size = embed_dim
        # self.decoder_block

        # Weight initialization
        for m in self.modules():
            if isinstance(m, (nn.Linear, nn.Conv2d)):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0.02)

            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0.02)
                nn.init.ones_(m.weight)

        nn.init.trunc_normal_(self.pos_embed, std=0.02)
        if self.pos_embed_win is not None:
            nn.init.trunc_normal_(self.pos_embed_win, std=0.02)
        if isinstance(self.classifier, nn.Linear):
            self.classifier.weight.data.mul_(0.001)
            self.classifier.bias.data.mul_(0.001)

    def _get_pos_embed(self) -> torch.Tensor:
        if self.pos_embed_win is not None:
            # Reference interpolation adapted from:
            # https://github.com/facebookresearch/sam2/blob/main/sam2/modeling/backbones/hieradet.py#L273
            pos_embed_win = self.pos_embed_win.tile(self.mask_spatial_shape)
            pos_embed = F.interpolate(
                self.pos_embed,
                size=pos_embed_win.shape[-2:],
                mode="bicubic",
                antialias=True,
            )
            pos_embed = pos_embed + pos_embed_win
            pos_embed = pos_embed.flatten(2).transpose(1, 2)
        else:
            pos_embed = self.pos_embed

        return pos_embed

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)
        x = x + self._get_pos_embed()
        x = self.unroll(x)

        out = {}
        for idx, (name, module) in enumerate(self.body.named_children()):
            x = module(x)
            if name in self.return_stages:
                z = self.reroll(x, self.block_count[idx])
                out[name] = z.permute(0, 3, 1, 2).contiguous()

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad_(False)

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad_(False)

    def masked_encoding_omission(
        self,
        x: torch.Tensor,
        ids_keep: Optional[torch.Tensor] = None,
        return_all_features: bool = False,
        return_keys: Literal["all", "tokens", "embedding"] = "tokens",
    ) -> TokenOmissionResultType:
        torch._assert(return_all_features is False, "not supported")  # pylint: disable=protected-access
        if ids_keep is not None:
            num_windows = math.prod(self.mask_spatial_shape)

            # Mask already in opposite form
            mask = mask_from_indices(ids_keep, num_windows)
            mask = mask.bool()
            patch_mask = mask.view(x.shape[0], 1, *self.mask_spatial_shape)  # B, C, *mask_spatial_shape
        else:
            patch_mask = None

        x = self.stem(x, patch_mask)
        x = x + self._get_pos_embed()
        x = self.unroll(x)

        # Discard masked tokens
        if ids_keep is not None:
            x = x[mask[..., None].tile(1, self.mu_size, x.shape[2])].view(x.shape[0], -1, x.shape[-1])

        x = self.body(x)

        result: TokenOmissionResultType = {}
        if return_keys in ("all", "tokens"):
            result["tokens"] = x

        if return_keys in ("all", "embedding"):
            x = self.features(x)
            result["embedding"] = x

        return result

    def masked_encoding(self, x: torch.Tensor, mask: torch.Tensor) -> tuple[list[torch.Tensor], torch.Tensor]:
        # Binary mask: 1 is *keep*, 0 is *remove*
        # Note this is opposite to original MAE
        mask = 1 - mask

        # Un-shuffle to get the binary mask
        mask = mask.bool()
        patch_mask = mask.view(x.shape[0], 1, *self.mask_spatial_shape)  # B, C, *mask_spatial_shape

        x = self.stem(x, patch_mask)
        x = x + self._get_pos_embed()
        x = self.unroll(x)

        # Discard masked tokens
        x = x[mask[..., None].tile(1, self.mu_size, x.shape[2])].view(x.shape[0], -1, x.shape[-1])

        out = []
        for idx, (name, module) in enumerate(self.body.named_children()):
            x = module(x)
            if name in self.return_stages:
                out.append(self.reroll(x, self.block_count[idx], mask=mask))

        return (out, mask)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        x = x + self._get_pos_embed()
        x = self.unroll(x)
        x = self.body(x)

        return x

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        x = self.features(x)

        return x

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        if self.pos_embed_win is not None:
            global_pos_size = (new_size[0] // 2**4, new_size[1] // 2**4)
            with torch.no_grad():
                pos_embed = F.interpolate(
                    self.pos_embed,
                    size=global_pos_size,
                    mode="bicubic",
                    antialias=True,
                )

            self.pos_embed = nn.Parameter(pos_embed)

        else:
            with torch.no_grad():
                pos_embed = adjust_position_embedding(
                    self.pos_embed,
                    (old_size[0] // self.patch_stride[0], old_size[1] // self.patch_stride[1]),
                    (new_size[0] // self.patch_stride[0], new_size[1] // self.patch_stride[1]),
                    0,
                )

            self.pos_embed = nn.Parameter(pos_embed)

        # Re-init vars
        self.tokens_spatial_shape = [i // s for i, s in zip(new_size, self.patch_stride)]
        self.mask_spatial_shape = [i // s for i, s in zip(self.tokens_spatial_shape, self.mask_unit_size)]

        # Re-init rolls
        self.unroll = Unroll(new_size, self.patch_stride, [self.q_stride] * len(self.stage_ends[:-1]))
        self.reroll = Reroll(
            new_size,
            self.patch_stride,
            [self.q_stride] * len(self.stage_ends[:-1]),
            self.stage_ends,
            self.q_pool,
        )


registry.register_model_config(
    "hiera_tiny",
    Hiera,
    config={
        "depths": [1, 2, 7, 2],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_small",
    Hiera,
    config={
        "depths": [1, 2, 11, 2],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_base",
    Hiera,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_base_plus",
    Hiera,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 112,
        "num_heads": 2,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_large",
    Hiera,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 144,
        "num_heads": 2,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "hiera_huge",
    Hiera,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 256,
        "num_heads": 4,
        "abs_win_pos_embed": False,
        "drop_path_rate": 0.3,
    },
)

# AbsWin variant
registry.register_model_config(
    "hiera_abswin_tiny",
    Hiera,
    config={
        "depths": [1, 2, 7, 2],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_abswin_small",
    Hiera,
    config={
        "depths": [1, 2, 11, 2],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_abswin_base",
    Hiera,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 96,
        "num_heads": 1,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_abswin_base_plus",
    Hiera,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 112,
        "num_heads": 2,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_abswin_base_plus_ap",
    Hiera,
    config={
        "depths": [2, 3, 16, 3],
        "embed_dim": 112,
        "num_heads": 2,
        "abs_win_pos_embed": True,
        "attn_pool_head": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "hiera_abswin_large",
    Hiera,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 144,
        "num_heads": 2,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.2,
    },
)
registry.register_model_config(
    "hiera_abswin_huge",
    Hiera,
    config={
        "depths": [2, 6, 36, 4],
        "embed_dim": 256,
        "num_heads": 4,
        "abs_win_pos_embed": True,
        "drop_path_rate": 0.3,
    },
)

registry.register_weights(
    "hiera_abswin_base_mim",
    {
        "url": "https://huggingface.co/birder-project/hiera_abswin_base_mim/resolve/main",
        "description": (
            "Hiera base with abswin image encoder pre-trained using Masked Image Modeling (MIM) for 400 epochs. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (224, 224),
        "formats": {
            "pt": {
                "file_size": 192.7,
                "sha256": "7133b4f857cca69d320173808d5c9921c1f05c129a76d58310a22bfafcd6deb1",
            },
            "safetensors": {
                "file_size": 192.6,
                "sha256": "48adf5f778081ea2e5aa98e10149d3bf8f1d4ad15e7998cb5611cd450557dfc1",
            },
        },
        "net": {"network": "hiera_abswin_base", "tag": "mim"},
    },
)
registry.register_weights(
    "hiera_abswin_base_mim-intermediate-eu-common",
    {
        "url": "https://huggingface.co/birder-project/hiera_abswin_base_mim-intermediate-eu-common/resolve/main",
        "description": (
            "Hiera base with abswin model with MIM pretraining and intermediate training, "
            "then fine-tuned on the eu-common dataset"
        ),
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 194.9,
                "sha256": "ce44e29c36b083c9e1d0eed9693a3f9c5bc39c03cf03efce3d453bfa328c7b9c",
            }
        },
        "net": {"network": "hiera_abswin_base", "tag": "mim-intermediate-eu-common"},
    },
)



import argparse

from birder.common import cli
from birder.eval import adversarial
from birder.eval import classification
from birder.eval.benchmarks import awa2
from birder.eval.benchmarks import bioscan5m
from birder.eval.benchmarks import fishnet
from birder.eval.benchmarks import flowers102
from birder.eval.benchmarks import fungiclef
from birder.eval.benchmarks import nabirds
from birder.eval.benchmarks import newt
from birder.eval.benchmarks import plankton
from birder.eval.benchmarks import plantdoc
from birder.eval.benchmarks import plantnet


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="python -m birder.eval",
        allow_abbrev=False,
        description="Evaluation module",
        epilog=(
            "Usage examples:\n"
            "python -m birder.eval adversarial -n resnet_v2_50 -t il-all -e 100 --method pgd "
            "--gpu data/validation_il-all_packed\n"
            "python -m birder.eval classification --filter '*il-all*' --gpu data/validation_il-all\n"
            "---\n"
            "python -m birder.eval awa2 --embeddings "
            "results/awa2/*.parquet --dataset-path ~/Datasets/Animals_with_Attributes2 --gpu\n"
            "python -m birder.eval bioscan5m --embeddings "
            "results/bioscan5m/*.parquet --data-path ~/Datasets/BIOSCAN-5M/species/testing_unseen\n"
            "python -m birder.eval fishnet --embeddings "
            "results/vit_b16_224px_embeddings.parquet --dataset-path ~/Datasets/fishnet --gpu\n"
            "python -m birder.eval flowers102 --embeddings "
            "results/flowers102_rope_i_vit_s16_pn_aps_c1_pe-core_0_384px_crop1.0_8189_sc_embeddings.parquet "
            "--dataset-path ~/Datasets/Flowers102\n"
            "python -m birder.eval fungiclef --embeddings "
            "results/fungiclef/*.parquet --dataset-path ~/Datasets/FungiCLEF2023 --gpu\n"
            "python -m birder.eval nabirds --embeddings "
            "results/vit_b16_224px_crop1.0_48562_embeddings.parquet --dataset-path ~/Datasets/nabirds --gpu\n"
            "python -m birder.eval newt --embeddings "
            "results/vit_reg4_so150m_p14_ls_dino-v2-bio_0_e45_224px_crop1.0_36032_output.parquet "
            "--dataset-path ~/Datasets/NeWT\n"
            "python -m birder.eval plankton --embeddings "
            "results/plankton/*.parquet --dataset-path ~/Datasets/plankton --gpu\n"
            "python -m birder.eval plantdoc --embeddings "
            "results/plantdoc_embeddings.parquet --dataset-path ~/Datasets/PlantDoc\n"
            "python -m birder.eval plantnet --embeddings "
            "results/plantnet_embeddings.parquet --dataset-path ~/Datasets/plantnet_300K\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparsers = parser.add_subparsers(dest="cmd", required=True)
    adversarial.set_parser(subparsers)
    classification.set_parser(subparsers)

    awa2.set_parser(subparsers)
    bioscan5m.set_parser(subparsers)
    fishnet.set_parser(subparsers)
    flowers102.set_parser(subparsers)
    fungiclef.set_parser(subparsers)
    nabirds.set_parser(subparsers)
    newt.set_parser(subparsers)
    plankton.set_parser(subparsers)
    plantdoc.set_parser(subparsers)
    plantnet.set_parser(subparsers)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



import logging
import unittest

import numpy as np
import torch
from PIL import Image
from torch import nn

from birder.introspection import base
from birder.introspection.attention_rollout import AttentionRollout
from birder.introspection.feature_pca import FeaturePCA
from birder.introspection.gradcam import GradCAM
from birder.introspection.guided_backprop import GuidedBackprop
from birder.introspection.transformer_attribution import TransformerAttribution
from birder.model_registry import registry

logging.disable(logging.CRITICAL)


class _TinyCNN(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(3, 8, 3, padding=1)
        self.relu = nn.ReLU()
        self.conv2 = nn.Conv2d(8, 16, 3, padding=1)
        self.flatten = nn.Flatten()
        self.fc = nn.Linear(16 * 16 * 16, 2)

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.relu(self.conv1(x))
        x = self.relu(self.conv2(x))
        return x

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        stage1 = self.relu(self.conv1(x))
        stage2 = self.relu(self.conv2(stage1))
        return {"stage1": stage1, "stage2": stage2}

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        x = self.flatten(x)
        return self.fc(x)


class TestIntrospectionBase(unittest.TestCase):
    def test_show_mask_on_image(self) -> None:
        img = np.random.rand(16, 16, 3).astype(np.float32)
        mask = np.random.rand(16, 16).astype(np.float32)

        result = base.show_mask_on_image(img, mask, image_weight=0.5)

        self.assertEqual(result.shape, (16, 16, 3))
        self.assertEqual(result.dtype, np.uint8)
        self.assertTrue(np.all(result >= 0))
        self.assertTrue(np.all(result <= 255))

    def test_scale_cam_image(self) -> None:
        cam = np.random.rand(2, 8, 8).astype(np.float32)

        result = base.scale_cam_image(cam, target_size=None)

        self.assertEqual(result.shape, (2, 8, 8))
        self.assertTrue(np.all(result >= 0))
        self.assertTrue(np.all(result <= 1))

        result_resized = base.scale_cam_image(cam, target_size=(16, 16))

        self.assertEqual(result_resized.shape, (2, 16, 16))

    def test_deprocess_image(self) -> None:
        img = np.random.randn(16, 16, 3).astype(np.float32)

        result = base.deprocess_image(img)

        self.assertEqual(result.shape, (16, 16, 3))
        self.assertEqual(result.dtype, np.uint8)
        self.assertTrue(np.all(result >= 0))
        self.assertTrue(np.all(result <= 255))

    def test_validate_target_class(self) -> None:
        # Valid target
        base.validate_target_class(0, num_classes=2)
        base.validate_target_class(1, num_classes=2)

        # None is valid
        base.validate_target_class(None, num_classes=2)

        # Invalid: out of range
        with self.assertRaises(ValueError):
            base.validate_target_class(2, num_classes=2)

        with self.assertRaises(ValueError):
            base.validate_target_class(-1, num_classes=2)

    def test_predict_class(self) -> None:
        logits = torch.tensor([[1.0, 2.0, 0.5]])
        pred = base.predict_class(logits)
        self.assertEqual(pred, 1)

        logits = torch.tensor([[2.0, 1.0]])
        pred = base.predict_class(logits)
        self.assertEqual(pred, 0)

    def test_preprocess_image(self) -> None:
        # Create test image
        img = Image.new("RGB", (32, 32), color=(255, 0, 0))

        def simple_transform(_x: Image.Image) -> torch.Tensor:
            return torch.rand(3, 16, 16)

        device = torch.device("cpu")
        input_tensor, rgb_img = base.preprocess_image(img, simple_transform, device)

        self.assertEqual(input_tensor.shape, (1, 3, 16, 16))
        self.assertEqual(rgb_img.shape, (16, 16, 3))
        self.assertTrue(np.all(rgb_img >= 0))
        self.assertTrue(np.all(rgb_img <= 1))


class TestInterpreters(unittest.TestCase):
    def setUp(self) -> None:
        torch.manual_seed(0)
        self.device = torch.device("cpu")

        # Create test image
        self.test_image = Image.new("RGB", (16, 16), color=(128, 128, 128))

        def simple_transform(x: Image.Image) -> torch.Tensor:
            arr = np.array(x).astype(np.float32) / 255.0
            return torch.from_numpy(arr).permute(2, 0, 1)

        self.transform = simple_transform

    def test_attention_rollout_result_structure(self) -> None:
        net = registry.net_factory("vit_t16", 2, size=(160, 160))

        # Create transform that resizes to match model input
        def vit_transform(x: Image.Image) -> torch.Tensor:
            x = x.resize((160, 160))
            arr = np.array(x).astype(np.float32) / 255.0
            return torch.from_numpy(arr).permute(2, 0, 1)

        interpreter = AttentionRollout(
            net, self.device, vit_transform, attention_layer_name="attn", discard_ratio=0.9, head_fusion="max"
        )
        result = interpreter(self.test_image, target_class=None)

        # Check result structure
        self.assertIsInstance(result.original_image, np.ndarray)
        self.assertIsInstance(result.visualization, np.ndarray)
        self.assertIsInstance(result.raw_output, np.ndarray)
        self.assertIsInstance(result.logits, torch.Tensor)
        self.assertIsInstance(result.predicted_class, int)

        # Check shapes
        self.assertEqual(len(result.original_image.shape), 3)
        self.assertEqual(len(result.visualization.shape), 3)
        self.assertEqual(result.logits.shape[-1], 2)  # type: ignore[union-attr]
        self.assertIn(result.predicted_class, [0, 1])

    def test_attention_rollout_no_duplicate_attentions(self) -> None:
        net = registry.net_factory("vit_t16", 2, size=(160, 160))

        # Create transform that resizes to match model input
        def vit_transform(x: Image.Image) -> torch.Tensor:
            x = x.resize((160, 160))
            arr = np.array(x).astype(np.float32) / 255.0
            return torch.from_numpy(arr).permute(2, 0, 1)

        # Track forward pass count
        original_forward = net.forward
        forward_count = {"count": 0}

        def counting_forward(x: torch.Tensor) -> torch.Tensor:
            forward_count["count"] += 1
            return original_forward(x)

        net.forward = counting_forward  # type: ignore[method-assign]

        interpreter = AttentionRollout(
            net, self.device, vit_transform, attention_layer_name="attn", discard_ratio=0.9, head_fusion="max"
        )

        # Track attention list length during execution
        attention_gatherer = interpreter.attention_gatherer
        original_call = attention_gatherer.__class__.__call__

        attention_lengths = []

        def tracking_call(self: object, x: torch.Tensor) -> tuple[list[torch.Tensor], torch.Tensor]:
            result = original_call(self, x)  # type: ignore[arg-type]
            attention_lengths.append(len(result[0]))
            return result

        attention_gatherer.__class__.__call__ = tracking_call  # type: ignore[method-assign]

        # Run interpreter
        _ = interpreter(self.test_image, target_class=None)

        # Restore original methods
        attention_gatherer.__class__.__call__ = original_call  # type: ignore[method-assign]
        net.forward = original_forward  # type: ignore[method-assign]

        # Verify only one forward pass occurred
        self.assertEqual(forward_count["count"], 1, "Should only perform ONE forward pass, not two!")

        # Verify attention list wasn't polluted
        self.assertEqual(len(attention_lengths), 1, "Attention gatherer should be called exactly once")
        num_encoder_layers = len([m for m in net.modules() if hasattr(m, "attn")])
        self.assertEqual(
            attention_lengths[0],
            num_encoder_layers,
            f"Should have {num_encoder_layers} attention maps, one per encoder layer",
        )

    def test_feature_pca_result_structure(self) -> None:
        net = _TinyCNN()

        interpreter = FeaturePCA(net, self.device, self.transform)
        result = interpreter(self.test_image)

        self.assertIsInstance(result.original_image, np.ndarray)
        self.assertIsInstance(result.visualization, np.ndarray)
        self.assertIsInstance(result.raw_output, np.ndarray)
        self.assertIsNone(result.logits)
        self.assertIsNone(result.predicted_class)

        self.assertEqual(len(result.original_image.shape), 3)
        self.assertEqual(len(result.visualization.shape), 3)
        self.assertEqual(result.visualization.shape[-1], 3)  # RGB channels
        self.assertEqual(result.visualization.dtype, np.uint8)

        self.assertEqual(len(result.raw_output.shape), 3)
        self.assertEqual(result.raw_output.shape[-1], 3)
        self.assertEqual(result.raw_output.dtype, np.float32)

    def test_feature_pca_values_normalized(self) -> None:
        net = _TinyCNN()

        interpreter = FeaturePCA(net, self.device, self.transform)
        result = interpreter(self.test_image)

        self.assertTrue(np.all(result.raw_output >= 0))
        self.assertTrue(np.all(result.raw_output <= 1))

        self.assertTrue(np.all(result.visualization >= 0))
        self.assertTrue(np.all(result.visualization <= 255))

    def test_gradcam_result_structure(self) -> None:
        net = _TinyCNN()
        target_layer = net.conv2

        interpreter = GradCAM(net, self.device, self.transform, target_layer)
        result = interpreter(self.test_image, target_class=None)

        # Check result structure
        self.assertIsInstance(result.original_image, np.ndarray)
        self.assertIsInstance(result.visualization, np.ndarray)
        self.assertIsInstance(result.raw_output, np.ndarray)
        self.assertIsInstance(result.logits, torch.Tensor)
        self.assertIsInstance(result.predicted_class, int)

        # Check shapes
        self.assertEqual(len(result.original_image.shape), 3)
        self.assertEqual(len(result.visualization.shape), 3)
        self.assertEqual(result.logits.shape[-1], 2)  # type: ignore[union-attr]
        self.assertIn(result.predicted_class, [0, 1])

    def test_gradcam_with_target_class(self) -> None:
        net = _TinyCNN()
        target_layer = net.conv2

        interpreter = GradCAM(net, self.device, self.transform, target_layer)
        result = interpreter(self.test_image, target_class=1)

        self.assertEqual(result.predicted_class, 1)

    def test_gradcam_invalid_target_class(self) -> None:
        net = _TinyCNN()
        target_layer = net.conv2

        interpreter = GradCAM(net, self.device, self.transform, target_layer)

        with self.assertRaises(ValueError):
            interpreter(self.test_image, target_class=5)

        with self.assertRaises(ValueError):
            interpreter(self.test_image, target_class=-1)

    def test_guided_backprop_result_structure(self) -> None:
        net = _TinyCNN()

        interpreter = GuidedBackprop(net, self.device, self.transform)
        result = interpreter(self.test_image, target_class=None)

        # Check result structure
        self.assertIsInstance(result.original_image, np.ndarray)
        self.assertIsInstance(result.visualization, np.ndarray)
        self.assertIsInstance(result.raw_output, np.ndarray)
        self.assertIsInstance(result.logits, torch.Tensor)
        self.assertIsInstance(result.predicted_class, int)

        # Check visualization is uint8
        self.assertEqual(result.visualization.dtype, np.uint8)
        self.assertTrue(np.all(result.visualization >= 0))
        self.assertTrue(np.all(result.visualization <= 255))

    def test_guided_backprop_model_restoration(self) -> None:
        net = _TinyCNN()
        original_relu_count = sum(1 for m in net.modules() if isinstance(m, nn.ReLU))

        interpreter = GuidedBackprop(net, self.device, self.transform)
        _ = interpreter(self.test_image, target_class=0)

        # Check model is restored
        restored_relu_count = sum(1 for m in net.modules() if isinstance(m, nn.ReLU))
        self.assertEqual(original_relu_count, restored_relu_count)

    def test_transformer_attribution_result_structure(self) -> None:
        net = registry.net_factory("vit_t16", 2, size=(160, 160))

        def vit_transform(x: Image.Image) -> torch.Tensor:
            x = x.resize((160, 160))
            arr = np.array(x).astype(np.float32) / 255.0
            return torch.from_numpy(arr).permute(2, 0, 1)

        interpreter = TransformerAttribution(net, self.device, vit_transform, attention_layer_name="attn")
        result = interpreter(self.test_image, target_class=None)

        self.assertIsInstance(result.original_image, np.ndarray)
        self.assertIsInstance(result.visualization, np.ndarray)
        self.assertIsInstance(result.raw_output, np.ndarray)
        self.assertIsInstance(result.logits, torch.Tensor)
        self.assertIsInstance(result.predicted_class, int)

        self.assertEqual(len(result.original_image.shape), 3)
        self.assertEqual(len(result.visualization.shape), 3)
        self.assertEqual(result.logits.shape[-1], 2)  # type: ignore[union-attr]
        self.assertIn(result.predicted_class, [0, 1])

    def test_transformer_attribution_with_target_class(self) -> None:
        net = registry.net_factory("vit_t16", 2, size=(160, 160))

        def vit_transform(x: Image.Image) -> torch.Tensor:
            x = x.resize((160, 160))
            arr = np.array(x).astype(np.float32) / 255.0
            return torch.from_numpy(arr).permute(2, 0, 1)

        interpreter = TransformerAttribution(net, self.device, vit_transform, attention_layer_name="attn")
        result = interpreter(self.test_image, target_class=1)

        self.assertEqual(result.predicted_class, 1)



"""
DeepFool

Paper "DeepFool: a simple and accurate method to fool deep neural networks", https://arxiv.org/abs/1511.04599

Generated by gpt-5.2-codex xhigh.
"""

from typing import Optional

import torch
from torch import nn

from birder.adversarial.base import AttackResult
from birder.adversarial.base import attack_success
from birder.adversarial.base import clamp_normalized
from birder.adversarial.base import predict_labels
from birder.adversarial.base import validate_target
from birder.data.transforms.classification import RGBType

GRAD_EPS = 1e-12


class DeepFool:
    def __init__(
        self, net: nn.Module, num_classes: int = 10, overshoot: float = 0.02, max_iter: int = 50, *, rgb_stats: RGBType
    ) -> None:
        if num_classes < 2:
            raise ValueError("num_classes must be at least 2")
        if max_iter <= 0:
            raise ValueError("max_iter must be positive")
        if overshoot < 0:
            raise ValueError("overshoot must be non-negative")

        self.net = net.eval()
        self.num_classes = num_classes
        self.overshoot = overshoot
        self.max_iter = max_iter
        self.rgb_stats = rgb_stats

    def __call__(self, input_tensor: torch.Tensor, target: Optional[torch.Tensor]) -> AttackResult:
        inputs = input_tensor.detach()
        with torch.no_grad():
            logits = self.net(inputs)

        target_labels = (
            validate_target(target, inputs.shape[0], logits.shape[1], inputs.device) if target is not None else None
        )
        targeted = target_labels is not None

        adv_inputs_list = []
        for idx in range(inputs.size(0)):
            target_label = target_labels[idx : idx + 1] if target_labels is not None else None
            adv_input = self._attack_single(inputs[idx : idx + 1], logits[idx : idx + 1], target_label)
            adv_inputs_list.append(adv_input)

        adv_inputs = torch.concat(adv_inputs_list, dim=0)
        with torch.no_grad():
            adv_logits = self.net(adv_inputs)

        success = attack_success(
            logits,
            adv_logits,
            targeted,
            target=target_labels if targeted else None,
        )

        return AttackResult(
            adv_inputs=adv_inputs,
            adv_logits=adv_logits,
            perturbation=adv_inputs - inputs,
            logits=logits.detach(),
            success=success,
        )

    def _attack_single(
        self, inputs: torch.Tensor, logits: torch.Tensor, target_label: Optional[torch.Tensor]
    ) -> torch.Tensor:
        adv_inputs = inputs.clone()
        original_label = int(predict_labels(logits).item())
        targeted = target_label is not None
        for _ in range(self.max_iter):
            adv_inputs.requires_grad_(True)
            outputs = self.net(adv_inputs)
            current_label = int(predict_labels(outputs).item())

            if targeted is True:
                assert target_label is not None
                target_value = int(target_label.item())
                if current_label == target_value:
                    break

                perturbation = self._targeted_perturbation(adv_inputs, outputs, current_label, target_value)

            else:
                if current_label != original_label:
                    break

                perturbation = self._untargeted_perturbation(adv_inputs, outputs, current_label)

            if perturbation is None:
                break

            # Overshoot helps ensure boundary crossing
            adv_inputs = adv_inputs.detach() + (1.0 + self.overshoot) * perturbation
            adv_inputs = clamp_normalized(adv_inputs, self.rgb_stats)

        return adv_inputs.detach()

    def _targeted_perturbation(
        self, adv_inputs: torch.Tensor, outputs: torch.Tensor, current_label: int, target_label: int
    ) -> Optional[torch.Tensor]:
        self.net.zero_grad(set_to_none=True)
        grad_current = torch.autograd.grad(outputs[0, current_label], adv_inputs, retain_graph=True)[0]
        grad_target = torch.autograd.grad(outputs[0, target_label], adv_inputs, retain_graph=False)[0]

        # Direction toward the target boundary
        w = grad_target - grad_current
        w_norm = torch.norm(w.view(-1))
        if w_norm.item() < GRAD_EPS:
            return None

        # Distance to the decision boundary
        f = outputs[0, target_label] - outputs[0, current_label]
        perturbation = (f.abs() / (w_norm**2 + GRAD_EPS)) * w

        return perturbation

    def _untargeted_perturbation(
        self, adv_inputs: torch.Tensor, outputs: torch.Tensor, current_label: int
    ) -> Optional[torch.Tensor]:
        # Search the top-k competing classes
        top_k = min(self.num_classes, outputs.shape[1])
        top_indices = torch.topk(outputs, k=top_k, dim=1).indices[0]
        candidate_labels = [int(idx) for idx in top_indices if int(idx) != current_label]

        if len(candidate_labels) == 0:
            return None

        self.net.zero_grad(set_to_none=True)
        grad_current = torch.autograd.grad(outputs[0, current_label], adv_inputs, retain_graph=True)[0]

        # Track the closest decision boundary
        best_dist = None
        best_w = None
        best_f = None
        for idx, label in enumerate(candidate_labels):
            # Keep the graph until the last class
            retain_graph = idx != len(candidate_labels) - 1
            grad_other = torch.autograd.grad(outputs[0, label], adv_inputs, retain_graph=retain_graph)[0]

            w_k = grad_other - grad_current
            w_norm = torch.norm(w_k.view(-1))
            if w_norm.item() < GRAD_EPS:
                continue

            f_k = outputs[0, label] - outputs[0, current_label]
            dist = f_k.abs() / (w_norm + GRAD_EPS)

            if best_dist is None or dist < best_dist:
                best_dist = dist
                best_w = w_k
                best_f = f_k

        if best_w is None or best_f is None:
            return None

        # Minimal perturbation toward the closest boundary
        best_w_norm = torch.norm(best_w.view(-1))
        if best_w_norm.item() < GRAD_EPS:
            return None

        perturbation = (best_f.abs() / (best_w_norm**2 + GRAD_EPS)) * best_w

        return perturbation



import logging
from collections.abc import Callable
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torchvision import tv_tensors
from torchvision.datasets import CocoDetection
from torchvision.datasets import wrap_dataset_for_transforms_v2
from torchvision.transforms.v2 import functional as F

from birder.data.transforms.mosaic import mosaic_fixed_grid
from birder.data.transforms.mosaic import mosaic_random_center

MosaicType = Literal["fixed_grid", "random_center"]

logger = logging.getLogger(__name__)


def _remove_images_without_annotations(dataset: CocoDetection, ignore_list: list[str]) -> CocoDetection:
    def _has_only_empty_bbox(anno: list[dict[str, Any]]) -> bool:
        return all(any(o <= 1 for o in obj["bbox"][2:]) for obj in anno)

    def _has_valid_annotation(anno: list[dict[str, Any]]) -> bool:
        # If it's empty, there is no annotation
        if len(anno) == 0:
            return False

        # If all boxes have close to zero area, there is no annotation
        if _has_only_empty_bbox(anno):
            return False

        return True

    ids = []
    for ds_idx, img_id in enumerate(dataset.ids):
        img_path = dataset.coco.loadImgs(img_id)[0]["file_name"]
        if img_path in ignore_list:
            logger.debug(f"Ignoring {img_path}")
            continue

        ann_ids = dataset.coco.getAnnIds(imgIds=img_id, iscrowd=None)
        anno = dataset.coco.loadAnns(ann_ids)
        if _has_valid_annotation(anno):
            ids.append(ds_idx)

    return torch.utils.data.Subset(dataset, ids)


def _convert_to_binary_annotations(dataset: CocoDetection) -> CocoDetection:
    for img_id in dataset.ids:
        ann_ids = dataset.coco.getAnnIds(imgIds=img_id, iscrowd=None)
        anno = dataset.coco.loadAnns(ann_ids)
        for obj in anno:
            obj["category_id"] = 1

    return dataset


class CocoBase(torch.utils.data.Dataset):
    def __init__(
        self, root: str | Path, ann_file: str, transforms: Optional[Callable[..., torch.Tensor]] = None
    ) -> None:
        super().__init__()
        dataset = CocoDetection(root, ann_file, transforms=transforms)
        self.class_to_idx = {cat["name"]: cat["id"] for cat in dataset.coco.cats.values()}

        # The transforms v2 wrapper causes open files count to "leak"
        # It seems due to the Pythonic COCO objects, maybe related to
        # https://ppwwyyxx.com/blog/2022/Demystify-RAM-Usage-in-Multiprocess-DataLoader/
        self.dataset = wrap_dataset_for_transforms_v2(dataset)

    def __getitem__(self, index: int) -> tuple[Any, ...]:
        raise NotImplementedError

    def __len__(self) -> int:
        return len(self.dataset)

    def __repr__(self) -> str:
        return repr(self.dataset)

    def remove_images_without_annotations(self, ignore_list: list[str]) -> None:
        self.dataset = _remove_images_without_annotations(self.dataset, ignore_list)

    def convert_to_binary_annotations(self) -> None:
        self.dataset = _convert_to_binary_annotations(self.dataset)
        self.class_to_idx = {"Object": 1}


class CocoTraining(CocoBase):
    def __getitem__(self, index: int) -> tuple[torch.Tensor, Any]:
        (sample, labels) = self.dataset[index]
        return (sample, labels)


class CocoInference(CocoBase):
    def __getitem__(self, index: int) -> tuple[str, torch.Tensor, Any, list[int]]:
        coco_id = self.dataset.ids[index]
        img_info = self.dataset.coco.loadImgs(coco_id)[0]
        path = img_info["file_name"]
        (sample, labels) = self.dataset[index]

        # Get original image size (height, width) before transforms
        orig_size = [img_info["height"], img_info["width"]]

        return (path, sample, labels, orig_size)


class CocoMosaicTraining(CocoBase):
    def __init__(
        self,
        root: str | Path,
        ann_file: str,
        transforms: Optional[Callable[..., torch.Tensor]],
        mosaic_transforms: Optional[Callable[..., torch.Tensor]],
        output_size: tuple[int, int],
        fill_value: int | tuple[int, int, int],
        mosaic_prob: float,
        mosaic_type: MosaicType = "fixed_grid",
    ) -> None:
        super().__init__(root, ann_file, transforms=None)
        self.transforms = transforms
        self.mosaic_transforms = mosaic_transforms
        self.output_size = output_size
        self.fill_value = fill_value
        self.mosaic_prob = mosaic_prob
        self.mosaic_fn = mosaic_fixed_grid if mosaic_type == "fixed_grid" else mosaic_random_center
        self._mosaic_base_prob: Optional[float] = None
        self._mosaic_decay_epochs: Optional[int] = None
        self._mosaic_decay_start: Optional[int] = None

    def configure_mosaic_linear_decay(self, base_prob: float, total_epochs: int, decay_fraction: float = 0.1) -> None:
        if total_epochs <= 0:
            raise ValueError("total_epochs must be positive")
        if decay_fraction <= 0.0 or decay_fraction > 1.0:
            raise ValueError("decay_fraction must be in (0.0, 1.0]")

        decay_epochs = max(1, int(total_epochs * decay_fraction))
        self._mosaic_base_prob = base_prob
        self._mosaic_decay_epochs = decay_epochs
        self._mosaic_decay_start = max(1, total_epochs - decay_epochs + 1)

    def update_mosaic_prob(self, epoch: int) -> Optional[float]:
        if self._mosaic_base_prob is None or self._mosaic_decay_epochs is None or self._mosaic_decay_start is None:
            return None

        if epoch >= self._mosaic_decay_start:
            if self._mosaic_decay_epochs <= 1:
                self.mosaic_prob = 0.0
            else:
                progress = (epoch - self._mosaic_decay_start) / (self._mosaic_decay_epochs - 1)
                self.mosaic_prob = max(0.0, self._mosaic_base_prob * (1 - progress))
        else:
            self.mosaic_prob = self._mosaic_base_prob

        return self.mosaic_prob

    @property
    def mosaic_decay_start(self) -> Optional[int]:
        return self._mosaic_decay_start

    def __getitem__(self, index: int) -> tuple[torch.Tensor, Any]:
        if torch.rand(1) < self.mosaic_prob:
            indices = [index] + torch.randint(len(self), (3,)).tolist()
            images = []
            targets = []
            for idx in indices:
                (img, target) = self.dataset[idx]
                images.append(img)
                targets.append(target)

            (img, target) = self.mosaic_fn(images, targets, self.output_size, self.fill_value)
            if self.mosaic_transforms is not None:
                (img, target) = self.mosaic_transforms(img, target)

        else:
            (img, target) = self.dataset[index]

            # When an image has no annotations, wrapped COCO returns only {'image_id': ...}
            if "boxes" not in target or len(target["boxes"]) == 0:
                canvas_size = F.get_size(img)
                boxes = tv_tensors.BoundingBoxes(
                    torch.zeros((0, 4), dtype=torch.float32),
                    format=tv_tensors.BoundingBoxFormat.XYXY,
                    canvas_size=canvas_size,
                )
                labels = torch.zeros((0,), dtype=torch.int64)
                target = {"boxes": boxes, "labels": labels}

            if self.transforms is not None:
                (img, target) = self.transforms(img, target)

        return (img, target)
